player = require 'player'

_TurnCount = 0
_HaveCornerSelected = false
_TargetCorner = 0
_TargetAngle = 0
_ArenaWidth = 0
_ArenaHeight = 0

local M = class(player, function(self, arenaWidth, arenaHeight)
	-- must init base!
	player.init(self)
    self.name = "Corner Pocket"
    self.R = 255
    self.G = 255
    self.B = 255
	
	_ArenaWidth = arenaWidth
	_ArenaHeight = arenaHeight
end)

-- data[1][1] = X
-- data[1][2] = Y
-- data[1][3] = bodyRotation
-- data[1][4] = turretRotation
-- data[1][5] = hitPoints
-- data[1][6] = turretX
-- data[1][7] = turretY
function M:process(data)
    -- do all calculations and setting of desired actions here
	_TurnCount = _TurnCount + 1
	if _HaveCornerSelected == false then
		-- calculate closest corner
		_HaveCornerSelected = true
		local bestCorner = 0
		local bestCornerDistance = 10000
		local bestCornerX = 0
		local bestCornerY = 0
		for corner = 1, 4, 1 do
			local tempDistance = distance(data[1][1], data[1][2], getCornerX(corner), getCornerY(corner))
			if tempDistance < bestCornerDistance then
				bestCorner = corner
				bestCornerDistance = tempDistance
				bestCornerX = getCornerX(corner)
				bestCornerY = getCornerY(corner)
			end
		end
		_TargetAngle = getRadian(data[1][1], data[1][2], bestCornerX, bestCornerY)
		if _TargetAngle > data[1][3] then
			self.bodyRotationDelta = .1
		else
			self.bodyRotationDelta = -.1
		end
	else
		if _TargetAngle == data[1][3] then
			self.bodyRotationDelta = 0
			_HaveCornerSelected = false
		elseif _TargetAngle > data[1][3] then
			local diffAngle = _TargetAngle - data[1][3]
			if diffAngle > .1 then
				self.bodyRotationDelta = .1
			else
				self.bodyRotationDelta = diffAngle
			end
		else
			local diffAngle = data[1][3] - _TargetAngle
			if diffAngle > -.1 then
				self.bodyRotationDelta = -.1
			else
				self.bodyRotationDelta = diffAngle
			end
		end
	end
	self.bodySpeed = 5
	-- fire as soon as I can
	if _TurnCount % 120 then
		self.shouldFire = true
	else
		self.shouldFire = false
	end
	-- Find the closest enemy tank
	-- First make a list of live enemy tanks
	local tankList = {}
	for iter = 2, #data, 1 do
		if data[iter][1] == true then
			table.insert(tankList, data[iter])
		end
	end
	-- Then find the closest
	local closestTank = {}
	local closestTankDistance = 10000
	for iter = 1, #tankList, 1 do
		local tempDistance = distance(data[1][1], data[1][2], tankList[iter][2], tankList[iter][3])
		if tempDistance < closestTankDistance then
			closestTank = tankList[iter]
			closestTankDistance = tempDistance
		end
	end
	-- Slew the gun to the closest tank
	local gunAngle = getRadian(data[1][6], data[1][7], closestTank[2], closestTank[3])
	if gunAngle > data[1][4] then
		if data[1][4] > 0 and data[1][4] < .5 * math.pi and gunAngle > 1.5 * math.pi and gunAngle < 2 * math.pi then
			if 2 * math.pi - gunAngle + data[1][4] >= .1 then
				self.turretDelta = -.1
			else
				self.turretDelta = -2 * math.pi - gunAngle + data[1][4]
			end
		
		elseif gunAngle - data[1][4] > .1 then
			self.turretDelta = .1
		else
			self.turretDelta = gunAngle - data[1][4]
		end
	elseif gunAngle < data[1][4] then
		if data[1][4] > 1.5 * math.pi and data[1][4] < 2 * math.pi and gunAngle > 0 and gunAngle < .5 * math.pi then
			if 2 * math.pi - data[1][4] + gunAngle >= .1 then
				self.turretDelta = .1
			else
				self.turretDelta = 2 * math.pi - data[1][4] + gunAngle
			end
		elseif data[1][4] - gunAngle > .1 then
			self.turretDelta = -.1
		else
			self.turretDelta = -1 * (data[1][4] - gunAngle)
		end
	else
		self.turretDelta = 0
	end
end

function invertAngle()
	if _TargetAngle >= math.pi then
		_TargetAngle = _TargetAngle - math.pi
	else
		_TargetAngle = _TargetAngle + math.pi
	end
end

function getCornerX(corner)
	if corner == 1 or corner == 3 then
		return 10 -- TODO 0 Magic numbers need to provide this value to the player
	else
		return _ArenaWidth
	end
end

function getCornerY(corner)
	if corner < 3 then
		return 10
	else
		return _ArenaHeight
	end
end

function distance(x, y, x2, y2)
	return math.sqrt(math.pow(x2 - x, 2) + math.pow(y2 - y, 2))
end

function getRadian(x, y, x2, y2)
	-- Due to Love2D orientation this is inverse and the second argument has been turned negative
	local theta = math.atan2((x2 - x), -1 * (y2 - y))
	-- The result is in the range [-π, π], the code below makes it positive
	if (theta < 0) then
		theta = theta + 2 * math.pi
	end
	return theta
end

return M
