player = require 'player'

local _TurnCount = 0
local _MaxX = 0
local _MaxY = 0
local _MinY = 0
local _MinX = 0
local WALL_LIMIT = 80
local _LastData = {}
local TURN_RATE = .1

local M = class(player, function(self, arenaMaxX, arenaMaxY, arenaMin)
	player.init(self) -- must init base!
    self.name = "Skim"

    self.R = 250
    self.G = 0
    self.B = 0
	
	_MinY = arenaMin
	_MinX = arenaMin
	_MaxX = arenaMaxX
	_MaxY = arenaMaxY
end)

-- data[1][1] = X
-- data[1][2] = Y
-- data[1][3] = bodyRotation
-- data[1][4] = turretRotation
-- data[1][5] = hitPoints
-- data[1][6] = turretX
-- data[1][7] = turretY
function M:process(data)
    -- do all calculations and setting of desired actions here
	_TurnCount = _TurnCount + 1
	
	-- TODO 0 if near a wall turn away otherwise turn toward the nearest wall
	-- Skim along the walls
	-- What is the nearest wall
	--   1
	-- 4   2
	--   3
	local closestWall = 0
	local closestWallDistance = 10000
	local tempDistance = 0
	local pointX = 0
	local pointY = 0
	for iter = 1, 4, 1 do
		if iter == 1 then
			pointX = data[1][1]
			pointY = _MinY
		elseif iter == 2 then
			pointX = _MaxX
			pointY = data[1][2]
		elseif iter == 3 then
			pointX = data[1][1]
			pointY = _MaxY
		else
			pointX = _MinX
			pointY = data[1][2]
		end
		tempDistance = distance(data[1][1], data[1][2], pointX, pointY)
		if tempDistance < closestWallDistance then
			closestWall = iter
			closestWallDistance = tempDistance
		end
	end
	-- Based on closest wall and direction, do the following:
	if closestWall == 4 and (data[1][3] >= math.pi * 1.5 or data[1][3] <= math.pi * .5) then
		-- Left, pointed up
		if closestWallDistance > WALL_LIMIT then
			self.bodyRotationDelta = -TURN_RATE
		else
			self.bodyRotationDelta = TURN_RATE
		end
	elseif closestWall == 2 and (data[1][3] >= math.pi * 1.5 or data[1][3] <= math.pi * .5) then
		-- Right, pointed up
		if closestWallDistance > WALL_LIMIT then
			self.bodyRotationDelta = TURN_RATE
		else
			self.bodyRotationDelta = -TURN_RATE
		end
	elseif closestWall == 4 and (data[1][3] <= math.pi * 1.5 and data[1][3] >= math.pi * .5) then
		-- Left, pointed down
		if closestWallDistance > WALL_LIMIT then
			self.bodyRotationDelta = TURN_RATE
		else
			self.bodyRotationDelta = -TURN_RATE
		end
	elseif closestWall == 2 and (data[1][3] <= math.pi * 1.5 and data[1][3] >= math.pi * .5) then
		-- Right, pointed down
		if closestWallDistance > WALL_LIMIT then
			self.bodyRotationDelta = -TURN_RATE
		else
			self.bodyRotationDelta = TURN_RATE
		end
	elseif closestWall == 1 and (data[1][3] >= 0 and data[1][3] <= math.pi) then
		-- Top, pointed right
		if closestWallDistance > WALL_LIMIT then
			self.bodyRotationDelta = -TURN_RATE
		else
			self.bodyRotationDelta = TURN_RATE
		end
	elseif closestWall == 3 and (data[1][3] >= 0 and data[1][3] <= math.pi) then
		-- Bottom, pointed right
		if closestWallDistance > WALL_LIMIT then
			self.bodyRotationDelta = TURN_RATE
		else
			self.bodyRotationDelta = -TURN_RATE
		end
	elseif closestWall == 1 and (data[1][3] >= math.pi and data[1][3] <= math.pi * 2) then
		-- Top, pointed left
		if closestWallDistance > WALL_LIMIT then
			self.bodyRotationDelta = TURN_RATE
		else
			self.bodyRotationDelta = -TURN_RATE
		end
	elseif closestWall == 3 and (data[1][3] >= math.pi and data[1][3] <= math.pi * 2) then
		-- Bottom, pointed left
		if closestWallDistance > WALL_LIMIT then
			self.bodyRotationDelta = -TURN_RATE
		else
			self.bodyRotationDelta = TURN_RATE
		end
	else
		-- NOTE: Should not be possible to get here
		self.bodyRotationDelta = 0
	end
	-- Turn when we hit a wall or a tank
	--if #_LastData > 0 then
	--	if data[1][1] == _LastData[1][1] and data[1][2] == _LastData[1][2] then
	--		self.bodyRotationDelta = .1
	--	else
	--		self.bodyRotationDelta = 0
	--	end
	--end
	_LastData = data
	-- Drive fast
	self.bodySpeed = 5
	-- Do not conserve firing
	self.shouldFire = true
	-- Find the closest enemy tank
	-- First make a list of live enemy tanks
	local tankList = {}
	for iter = 2, #data, 1 do
		if data[iter][1] == true then
			table.insert(tankList, data[iter])
		end
	end
	-- Then find the closest
	local closestTank = {}
	local closestTankDistance = 10000
	for iter = 1, #tankList, 1 do
		local tempDistance = distance(data[1][1], data[1][2], tankList[iter][2], tankList[iter][3])
		if tempDistance < closestTankDistance then
			closestTank = tankList[iter]
			closestTankDistance = tempDistance
		end
	end
	-- Slew the gun to the closest tank
	local gunAngle = getRadian(data[1][6], data[1][7], closestTank[2], closestTank[3])
	if gunAngle > data[1][4] then
		if data[1][4] > 0 and data[1][4] < .5 * math.pi and gunAngle > 1.5 * math.pi and gunAngle < 2 * math.pi then
			if 2 * math.pi - gunAngle + data[1][4] >= .1 then
				self.turretDelta = -.1
			else
				self.turretDelta = -2 * math.pi - gunAngle + data[1][4]
			end
		
		elseif gunAngle - data[1][4] > .1 then
			self.turretDelta = .1
		else
			self.turretDelta = gunAngle - data[1][4]
		end
	elseif gunAngle < data[1][4] then
		if data[1][4] > 1.5 * math.pi and data[1][4] < 2 * math.pi and gunAngle > 0 and gunAngle < .5 * math.pi then
			if 2 * math.pi - data[1][4] + gunAngle >= .1 then
				self.turretDelta = .1
			else
				self.turretDelta = 2 * math.pi - data[1][4] + gunAngle
			end
		elseif data[1][4] - gunAngle > .1 then
			self.turretDelta = -.1
		else
			self.turretDelta = -1 * (data[1][4] - gunAngle)
		end
	else
		self.turretDelta = 0
	end
end

function distance(x, y, x2, y2)
	return math.sqrt(math.pow(x2 - x, 2) + math.pow(y2 - y, 2))
end

function getRadian(x, y, x2, y2)
	-- Due to Love2D orientation this is inverse and the second argument has been turned negative
	local theta = math.atan2((x2 - x), -1 * (y2 - y))
	-- The result is in the range [-π, π], the code below makes it positive
	if (theta < 0) then
		theta = theta + 2 * math.pi
	end
	return theta
end

return M
